"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const BaseRoute_1 = require("./BaseRoute");
const tsioc_1 = require("tsioc");
const decorators_1 = require("../decorators");
const decorators_2 = require("../decorators");
const util_1 = require("../util");
const index_1 = require("../errors/index");
const util_2 = require("util");
const restults_1 = require("../restults");
const RequestMethod_1 = require("../RequestMethod");
const index_2 = require("../index");
class ControllerRoute extends BaseRoute_1.BaseRoute {
    constructor(route, controller) {
        super(route);
        this.controller = controller;
    }
    async options(container, ctx, next) {
        await this.invokeOption(ctx, container, next);
    }
    async navigating(container, ctx) {
        try {
            let decorator = this.getDecoratorByMethod(ctx.method);
            if (decorator !== decorators_1.Options) {
                await this.invoke(ctx, container, decorator, (meta, params, ctrl) => this.createProvider(container, meta, params, ctrl, ctx));
            }
            else {
                throw new index_1.BadRequestError();
            }
        }
        catch (err) {
            if (err instanceof index_1.HttpError) {
                ctx.status = err.code;
                ctx.response.status = err.code;
                ctx.response.message = err.message;
            }
            console.log(err.toString());
        }
    }
    async invokeOption(ctx, container, next) {
        let requestOrigin = ctx.get('Origin');
        ctx.vary('Origin');
        if (!requestOrigin) {
            return next();
        }
        let origin = requestOrigin;
        let headersSet = {};
        let set = (key, value) => {
            ctx.set(key, value);
            headersSet[key] = value;
        };
        let config = container.get(index_2.Configuration);
        let options = config.corsOptions || {};
        if (ctx.method !== 'OPTIONS') {
            set('Access-Control-Allow-Origin', origin);
            if (options.credentials === true) {
                set('Access-Control-Allow-Credentials', 'true');
            }
            if (options.exposeHeaders) {
                set('Access-Control-Expose-Headers', options.exposeHeaders);
            }
            if (!options.keepHeadersOnError) {
                return next();
            }
            return next().catch(err => {
                err.headers = Object.assign({}, err.headers, headersSet);
                throw err;
            });
        }
        else {
            if (!ctx.get('Access-Control-Request-Method')) {
                // this not preflight request, ignore it
                return next();
            }
            let methodCors = tsioc_1.getMethodMetadata(decorators_1.Cors, this.controller);
            let method = RequestMethod_1.parseRequestMethod(ctx.get('Access-Control-Request-Method'));
            let meta = this.getRouteMetaData(ctx, container, this.getDecoratorByMethod(method));
            if (meta && meta.propertyKey) {
                let corsmetas = tsioc_1.getMethodMetadata(decorators_1.Cors, this.controller)[meta.propertyKey] || [];
                if (corsmetas.length < 1) {
                    corsmetas = tsioc_1.getTypeMetadata(decorators_1.Cors, this.controller);
                }
                if (corsmetas.length) {
                    options = Object.assign({}, options, corsmetas[0]);
                    ctx.set('Access-Control-Allow-Origin', origin);
                    if (options.credentials === true) {
                        ctx.set('Access-Control-Allow-Credentials', 'true');
                    }
                    let maxAge = String(options.maxAge);
                    if (maxAge) {
                        ctx.set('Access-Control-Max-Age', maxAge);
                    }
                    let allowsM = util_2.isArray(options.allowMethods) ? options.allowMethods.map(m => RequestMethod_1.methodToString(m)).join(',') : options.allowMethods;
                    allowsM = allowsM || 'GET,HEAD,PUT,POST,DELETE,PATCH';
                    if (allowsM) {
                        ctx.set('Access-Control-Allow-Methods', allowsM);
                    }
                    let allowH = util_2.isArray(options.allowHeaders) ? options.allowHeaders.join(',') : options.allowHeaders;
                    if (!allowH) {
                        allowH = ctx.get('Access-Control-Request-Headers');
                    }
                    if (allowH) {
                        ctx.set('Access-Control-Allow-Headers', allowH);
                    }
                    ctx.status = 204;
                    return;
                }
            }
            return next();
        }
    }
    async invoke(ctx, container, decorator, provider) {
        let meta = this.getRouteMetaData(ctx, container, decorator);
        if (meta && meta.propertyKey) {
            let ctrl = container.get(this.controller);
            if (container.has(util_1.symbols.IAuthorization)) {
                let hasAuth = Reflect.hasMetadata(decorators_2.Authorization.toString(), ctrl, meta.propertyKey);
                if (hasAuth) {
                    let auth = container.get(util_1.symbols.IAuthorization);
                    if (!auth.isAuth()) {
                        throw new index_1.UnauthorizedError();
                    }
                }
            }
            let params = container.getMethodParameters(this.controller, ctrl, meta.propertyKey);
            let response = await container.invoke(this.controller, meta.propertyKey, ctrl, ...provider(meta, params, ctrl));
            let contentType = meta.contentType;
            if (util_2.isString(response)) {
                contentType = contentType || 'text/html';
                ctx.response.body = response;
            }
            else if (util_2.isArray(response) && response instanceof Uint8Array) {
                ctx.response.body = Buffer.from(response);
            }
            else if (util_2.isObject(response)) {
                if (response instanceof restults_1.ResultValue) {
                    await response.sendValue(ctx, container);
                }
                else {
                    contentType = contentType || 'application/json';
                    ctx.type = contentType;
                    ctx.response.body = response;
                    ctx.redirect;
                }
            }
        }
        else {
            throw new index_1.NotFoundError();
        }
    }
    getDecoratorByMethod(method) {
        if (util_2.isNumber(method)) {
            method = RequestMethod_1.methodToString(method);
        }
        else {
            method = (method || '').toLowerCase();
        }
        switch (method) {
            case 'GET':
                return decorators_1.Get;
            case 'POST':
                return decorators_1.Post;
            case 'Put':
                return decorators_1.Put;
            case 'DElETE':
                return decorators_1.Delete;
            case 'OPTIONS':
                return decorators_1.Options;
            default:
                return decorators_1.Get;
        }
    }
    createProvider(container, meta, params, ctrl, ctx) {
        if (params && params.length) {
            let paramVal = null;
            if (this.isRestUri(meta.route)) {
                let route = meta.route.substr(0, meta.route.indexOf('/:')) + '/';
                let baseURL = this.cutEmptyPath(this.url, true);
                let routeUrl = this.cutEmptyPath(ctx.url.replace(baseURL, ''));
                let querystring = routeUrl.replace(route, '');
                paramVal = querystring.indexOf('/') > 0 ? querystring.substr(0, querystring.indexOf('/')) : querystring;
            }
            let body = ctx.request['body'] || {};
            let providers = params.map((p, idx) => {
                try {
                    if (tsioc_1.isClass(p) && !this.isBaseType(p)) {
                        let meta = tsioc_1.getPropertyMetadata(decorators_1.Field, p);
                        let val = container.get(p);
                        for (let n in meta) {
                            if (!util_2.isUndefined(body[n])) {
                                val[n] = body[n];
                            }
                        }
                        return {
                            value: val,
                            index: idx
                        };
                    }
                    else if (this.isBaseType(p)) {
                        let val;
                        if (paramVal !== null) {
                            if (p === String) {
                                val = paramVal;
                            }
                            else if (p === Boolean) {
                                val = new Boolean(paramVal);
                            }
                            else if (p === Number) {
                                val = parseFloat(paramVal);
                            }
                            else if (p === Date) {
                                val = new Date(paramVal);
                            }
                            paramVal = null;
                        }
                        return {
                            value: val,
                            index: idx
                        };
                    }
                    else {
                        return null;
                    }
                }
                catch (err) {
                    throw new index_1.BadRequestError(err.toString());
                }
            });
            return providers.filter(p => p !== null);
        }
        return [];
    }
    isBaseType(p) {
        if (!tsioc_1.isToken(p)) {
            return true;
        }
        if (p === Boolean || p === String || p === Number) {
            return true;
        }
        if (!tsioc_1.isFunction(p)) {
            return true;
        }
        return false;
    }
    isRestUri(uri) {
        return /\/:/.test(uri || '');
    }
    getRouteMetaData(ctx, container, decorator) {
        let decoratorName = decorator.toString();
        let baseURL = this.cutEmptyPath(this.url, true);
        let routPath = this.cutEmptyPath(ctx.url.replace(baseURL, ''));
        let methodMaps = tsioc_1.getMethodMetadata(decoratorName, this.controller);
        let meta;
        let allMethods = [];
        for (let name in methodMaps) {
            allMethods = allMethods.concat(methodMaps[name]);
        }
        allMethods = allMethods.sort((ra, rb) => (rb.route || '').length - (ra.route || '').length);
        meta = allMethods.find(route => (route.route || '') === routPath);
        if (!meta) {
            meta = allMethods.find(route => {
                let uri = route.route || '';
                // if (uri === routPath) {
                //     return true;
                // }
                if (this.isRestUri(uri)) {
                    let idex = uri.indexOf('/:');
                    let url = uri.substr(0, idex);
                    if (url !== routPath && routPath.indexOf(url) === 0) {
                        return true;
                    }
                }
                return false;
            });
        }
        return meta;
    }
}
exports.ControllerRoute = ControllerRoute;

//# sourceMappingURL=../sourcemaps/router/ControllerRoute.js.map
